import configparser
import os
from typing import Optional

from rust2rpm import log


def to_list(s, sort=True):
    if not s:
        return []
    if sort:
        return list(sorted(filter(None, (l.strip() for l in s.splitlines()))))
    else:
        return list(filter(None, (l.strip() for l in s.splitlines())))


class Rust2RpmConfError(ValueError):
    pass


class Rust2RpmConf:
    def __init__(
        self,
        *,
        summary: Optional[str] = None,
        supported_arches: list[str] = None,
        all_features: bool = False,
        unwanted_features: list[str] = None,
        enabled_features: list[str] = None,
        buildrequires: list[str] = None,
        testrequires: list[str] = None,
        bin_requires: list[str] = None,
        lib_requires: dict[Optional[str], list[str]] = None,
    ):
        self.summary = summary
        self.supported_arches: list[str] = supported_arches or list()
        self.all_features: bool = all_features
        self.unwanted_features: list[str] = unwanted_features or list()
        self.enabled_features: list[str] = enabled_features or list()
        self.buildrequires: list[str] = buildrequires or list()
        self.testrequires: list[str] = testrequires or list()
        self.bin_requires: list[str] = bin_requires or list()
        self.lib_requires: dict[Optional[str], list[str]] = lib_requires or dict()

    def __eq__(self, other):
        if not isinstance(other, Rust2RpmConf):
            return False  # pragma nocover

        # order of items in the list is consistent:
        # lists are sorted when parsing rust2rpm.conf files
        return (
            self.summary == other.summary
            and self.supported_arches == other.supported_arches
            and self.all_features == other.all_features
            and self.unwanted_features == other.unwanted_features
            and self.enabled_features == other.enabled_features
            and self.buildrequires == other.buildrequires
            and self.testrequires == other.testrequires
            and self.bin_requires == other.bin_requires
            and self.lib_requires == other.lib_requires
        )

    @staticmethod
    def load(filenames: list[str], target: str, features: set[str]):
        conf = configparser.ConfigParser(interpolation=configparser.ExtendedInterpolation())
        confs = conf.read(filenames)

        if len(confs) > 1:  # pragma nocover
            raise FileExistsError

        # clean up configuration files with deprecated names
        if ".rust2rpm.conf" in confs:  # pragma nocover
            os.rename(".rust2rpm.conf", "rust2rpm.conf")
            log.info("Renamed deprecated, hidden .rust2rpm.conf file to rust2rpm.conf.")

        if "_rust2rpm.conf" in confs:  # pragma nocover
            os.rename("_rust2rpm.conf", "rust2rpm.conf")
            log.info("Renamed deprecated _rust2rpm.conf file to rust2rpm.conf.")

        # merge target-specific configuration with default configuration
        if target not in conf:
            conf.add_section(target)
        merged = conf[target]

        # validate configuration file
        valid_targets = ["fedora", "mageia", "opensuse", "plain"]
        valid_keys = [
            "summary",
            "supported-arches",
            "all-features",
            "unwanted-features",
            "enabled-features",
            "buildrequires",
            "testrequires",
            "lib.requires",
            "bin.requires",
        ]

        # the "default" feature is always implicitly defined
        if "default" not in features:
            features.add("default")

        for feature in features:
            valid_keys.append(f"lib+{feature}.requires")

        # check section names
        for section in conf.keys():
            if section != "DEFAULT" and section not in valid_targets:
                raise Rust2RpmConfError(f"Invalid section: {section!r}")

        # check setting keys
        for key in merged.keys():
            if key not in valid_keys:
                raise Rust2RpmConfError(f"Invalid key: {key!r}")

        # parse configuration and validate settings
        settings = dict()

        if summary := merged.get("summary"):
            settings["summary"] = summary

        if supported_arches := merged.get("supported-arches"):
            settings["supported_arches"] = to_list(supported_arches, False)

        all_features = merged.getboolean("all-features")
        if all_features is not None:
            settings["all_features"] = all_features

        if unwanted_features := merged.get("unwanted-features"):
            settings["unwanted_features"] = to_list(unwanted_features)

            for unwanted_feature in settings["unwanted_features"]:
                if unwanted_feature not in features:
                    raise Rust2RpmConfError(f'Unrecognized "unwanted" feature: {unwanted_feature!r}')

        if enabled_features := merged.get("enabled-features"):
            settings["enabled_features"] = to_list(enabled_features)

            for enabled_feature in settings["enabled_features"]:
                if enabled_feature not in features:
                    raise Rust2RpmConfError(f'Unrecognized "enabled" feature: {enabled_feature!r}')

        if buildrequires := merged.get("buildrequires"):
            settings["buildrequires"] = to_list(buildrequires)
        if testrequires := merged.get("testrequires"):
            settings["testrequires"] = to_list(testrequires)
        if bin_requires := merged.get("bin.requires"):
            settings["bin_requires"] = to_list(bin_requires)

        if lib_requires := merged.get("lib.requires"):
            if "lib_requires" not in settings.keys():
                settings["lib_requires"] = dict()
            settings["lib_requires"][None] = to_list(lib_requires)

        for feature in features:
            if lib_feature_requires := merged.get(f"lib+{feature}.requires"):
                if "lib_requires" not in settings.keys():
                    settings["lib_requires"] = dict()
                settings["lib_requires"][feature] = to_list(lib_feature_requires)

        return Rust2RpmConf(**settings)
