import os
import re
from typing import Optional
from urllib.parse import urljoin

import requests
import tqdm

from rust2rpm import log
from rust2rpm.utils import remove_on_error

CRATES_IO_API_URL = "https://crates.io/api/v1/"
XDG_CACHE_HOME = os.getenv("XDG_CACHE_HOME", os.path.expanduser("~/.cache"))
CACHE_DIR = os.path.join(XDG_CACHE_HOME, "rust2rpm")


class NoVersionsError(Exception):
    pass


def query_newest_version(crate: str) -> str:
    url = urljoin(CRATES_IO_API_URL, f"crates/{crate}/versions")
    req = requests.get(url, headers={"User-Agent": "rust2rpm"})
    req.raise_for_status()
    versions = req.json()["versions"]

    is_stable = lambda s: not re.search("alpha|beta|rc|pre", s["num"])
    is_not_yanked = lambda s: not s["yanked"]

    # return the most recent, non-yanked stable version

    is_not_yanked_and_stable = lambda s: is_stable(s) and is_not_yanked(s)
    not_yanked_and_stable = [*filter(is_not_yanked_and_stable, versions)]
    if len(not_yanked_and_stable) > 0:
        version = not_yanked_and_stable[0]["num"]
        return version

    # there are no non-yanked stable versions:
    # fall back to the latest pre-release

    not_yanked = [*filter(is_not_yanked, versions)]
    if len(not_yanked) > 0:
        version = not_yanked[0]["num"]
        log.warn(f"No stable versions available. Falling back to the latest pre-release version {version!r}.")
        return version

    # there are no non-yanked versions: fatal
    raise NoVersionsError()


def download_crate(name: str, version: Optional[str] = None) -> tuple[str, str]:
    if version is None:
        # query crates.io for the latest version
        version = query_newest_version(name)

    os.makedirs(CACHE_DIR, exist_ok=True)

    crate_file_name = f"{name}-{version}.crate"
    crate_file_path = os.path.join(CACHE_DIR, crate_file_name)

    if not os.path.isfile(crate_file_path):
        url = urljoin(CRATES_IO_API_URL, f"crates/{name}/{version}/download#")
        req = requests.get(url, stream=True, headers={"User-Agent": "rust2rpm"})
        req.raise_for_status()
        total = int(req.headers["Content-Length"])

        with remove_on_error(crate_file_path), open(crate_file_path, "wb") as f:
            for chunk in tqdm.tqdm(
                req.iter_content(),
                f"Downloading {crate_file_name}",
                total=total,
                unit="B",
                unit_scale=True,
            ):
                f.write(chunk)

    return crate_file_path, version
