"""
Functionality for rendering the spec file template.
"""

import time
from typing import Optional

from cargo2rpm.metadata import Metadata, Package, Version, FeatureFlags
from cargo2rpm import rpm
import jinja2

from rust2rpm import __version__, log
from rust2rpm.conf import Rust2RpmConf
from rust2rpm.licensing import translate_license
from rust2rpm.metadata import package_uses_rust_1_60_feature_syntax, get_required_features_for_binaries


def spec_file_template(filename: str):
    env = jinja2.Environment(
        loader=jinja2.ChoiceLoader([jinja2.FileSystemLoader(["/"]), jinja2.PackageLoader("rust2rpm", "templates")]),
        extensions=["jinja2.ext.do"],
        trim_blocks=True,
        lstrip_blocks=True,
    )

    return env.get_template(filename)


def template_args_fedora(date: Optional[time.struct_time], packager: Optional[str], rpmautospec: bool) -> dict:
    kwargs = dict()

    kwargs["include_build_requires"] = False
    kwargs["include_provides"] = False
    kwargs["include_requires"] = False

    if rpmautospec:
        kwargs["rpm_release"] = "%autorelease"
    else:
        kwargs["rpm_release"] = "1%{?dist}"

    kwargs["rpm_changelog_date"] = time.strftime(*(["%a %b %d %Y"] + ([date] if date else [])))
    kwargs["rpm_changelog_packager"] = packager or "rust2rpm <nobody@fedoraproject.org>"

    return kwargs


def template_args_mageia(date: Optional[time.struct_time], packager: Optional[str]) -> dict:
    kwargs = dict()

    kwargs["include_build_requires"] = True
    kwargs["include_provides"] = False
    kwargs["include_requires"] = False

    kwargs["rpm_release"] = "%mkrel 1"
    kwargs["rpm_group"] = "Development/Rust"

    kwargs["rpm_changelog_date"] = time.strftime(*(["%a %b %d %Y"] + ([date] if date else [])))
    kwargs["rpm_changelog_packager"] = packager or "rust2rpm <nobody@mageia.org>"

    return kwargs


def template_args_opensuse(date: Optional[time.struct_time], packager: Optional[str]) -> dict:
    kwargs = dict()

    kwargs["include_build_requires"] = True
    kwargs["include_provides"] = False
    kwargs["include_requires"] = False

    kwargs["spec_copyright_year"] = time.strftime("%Y")
    kwargs["rpm_release"] = "0"
    kwargs["rpm_group"] = "Development/Libraries/Rust"

    kwargs["rpm_changelog_date"] = time.strftime(*(["%a %b %d %T %Z %Y"] + ([date] if date else [])))
    kwargs["rpm_changelog_packager"] = packager or "rust2rpm <opensuse-packaging@opensuse.org>"

    return kwargs


def template_args_plain(date: Optional[time.struct_time], packager: Optional[str]) -> dict:
    kwargs = dict()

    kwargs["include_build_requires"] = True
    kwargs["include_provides"] = True
    kwargs["include_requires"] = True

    kwargs["rpm_release"] = "1%{?dist}"

    kwargs["rpm_changelog_date"] = time.strftime(*(["%a %b %d %Y"] + ([date] if date else [])))
    kwargs["rpm_changelog_packager"] = packager or "rust2rpm <packager@example.com>"

    return kwargs


def spec_render_crate(
    *,
    metadata: Metadata,
    target: str,
    rpm_name: str,
    patch_file_automatic: Optional[str],
    patch_file_manual: Optional[str],
    license_files: list[str],
    doc_files: list[str],
    distconf: Rust2RpmConf,
    feature_flags: FeatureFlags,
    relative_license_paths: bool,
    rpmautospec: bool,
    auto_changelog_entry: bool,
    date: Optional[time.struct_time] = None,
    packager: Optional[str],
):
    template = spec_file_template("crate.spec")

    # enforce consistent ordering of binaries
    binaries = [*metadata.get_binaries()]
    binaries.sort()

    is_bin = metadata.is_bin()
    is_lib = metadata.is_lib()
    is_cdylib = metadata.is_cdylib()

    package = metadata.packages[0]
    description = package.get_description()
    summary = package.get_summary()

    # enforce consistent ordering of feature subpackages
    features = list(sorted(package.get_feature_names()))
    if "default" in features:
        features.remove("default")
    features.insert(0, None)
    features.insert(1, "default")

    generator_version = __version__.split(".")[0]

    buildrequires = rpm.buildrequires(package, feature_flags, False)
    test_requires = set.difference(
        rpm.buildrequires(package, feature_flags, True), rpm.buildrequires(package, feature_flags, False)
    )

    rpm_buildrequires = list(sorted(buildrequires))
    rpm_test_requires = list(sorted(test_requires))
    rpm_requires = {feature: list(sorted(rpm.requires(package, feature))) for feature in features}
    rpm_provides = {feature: rpm.provides(package, feature) for feature in features}

    conf_lib_requires = dict()
    for feature in features:
        if feature is None:
            conf_key = "lib"
        else:
            conf_key = f"lib+{feature}"
        conf_lib_requires[conf_key] = distconf.lib_requires.get(feature) or list()

    for feature in distconf.unwanted_features:
        features.remove(feature)

    if package_uses_rust_1_60_feature_syntax(package.features):
        rust_packaging_dep = "cargo-rpm-macros >= 24"
    elif is_bin or is_cdylib:
        rust_packaging_dep = "rust-packaging >= 23"
    else:
        rust_packaging_dep = "rust-packaging >= 21"

    if feature_flags.all_features or distconf.all_features:
        cargo_args = " -a"
    elif required_features := get_required_features_for_binaries(package):
        # remove required features that are already part of the enabled feature set
        enabled_by_default = package.get_enabled_features_transitive(feature_flags)[0]
        for f in enabled_by_default:
            if f in required_features:
                required_features.remove(f)

        # merge, de-duplicate, and re-sort lists of enabled features
        enabled_features = sorted(set.union(set(required_features), set(distconf.enabled_features)))

        if len(enabled_features) > 0:
            cargo_args = f" -f {','.join(enabled_features)}"
        else:
            cargo_args = ""
    elif distconf.enabled_features:
        # list of features is already sorted when parsing the configuration file
        cargo_args = f" -f {','.join(distconf.enabled_features)}"
    else:
        cargo_args = ""

    if license_str := metadata.packages[0].license:
        rpm_license, rpm_license_comments = translate_license(target, license_str)
    else:
        log.warn("No license information in crate metadata.")
        rpm_license = None
        rpm_license_comments = "# FIXME: No license information in crate metadata."

    template_args_common = {
        # Parameters specific to rust2rpm
        "rust2rpm_version": generator_version,
        "rust2rpm_target": target,
        "rust_packaging_dep": rust_packaging_dep,
        # Parameters for RPM package metadata
        "rpm_name": rpm_name,
        "rpm_version": Version.parse(package.version).to_rpm(),
        "rpm_summary": distconf.summary or summary,
        "rpm_description": description,
        "rpm_license": rpm_license,
        "rpm_license_comments": rpm_license_comments,
        "rpm_patch_file_automatic": patch_file_automatic,
        "rpm_patch_file_manual": patch_file_manual,
        "rpm_buildrequires": rpm_buildrequires,
        "rpm_test_requires": rpm_test_requires,
        "rpm_requires": rpm_requires,
        "rpm_provides": rpm_provides,
        "rpm_license_files": license_files,
        "rpm_doc_files": doc_files,
        # Parameters that control generation of subpackages
        "rpm_binary_package": is_bin,
        "rpm_cdylib_package": is_cdylib,
        "rpm_library_package": is_lib,
        "rpm_binary_names": binaries,
        "crate_features": features,
        # Parameters for crate metadata
        "crate_name": package.name,
        "crate_version": package.version,
        "crate_license": package.license,
        # Parameters derived from rust2rpm.conf
        "conf_buildrequires": distconf.buildrequires,
        "conf_test_requires": distconf.testrequires,
        "conf_bin_requires": distconf.bin_requires,
        "conf_lib_requires": conf_lib_requires,
        "conf_supported_arches": " ".join(distconf.supported_arches),
        # Parameters derived from command-line flags
        "cargo_args": cargo_args,
        "use_relative_license_paths": relative_license_paths,
        "use_rpmautospec": rpmautospec,
        "make_changelog_entry": auto_changelog_entry,
    }

    match target:
        case "fedora":
            template_args_target = template_args_fedora(date, packager, rpmautospec)
        case "mageia":
            template_args_target = template_args_mageia(date, packager)
        case "opensuse":
            template_args_target = template_args_opensuse(date, packager)
        case "plain":
            template_args_target = template_args_plain(date, packager)
        case _:
            raise ValueError(f"Unknown target {target!r} (this should never happen)")

    spec_contents = template.render(
        **template_args_common,
        **template_args_target,
    )

    if not spec_contents.endswith("\n"):
        spec_contents += "\n"

    return spec_contents


def spec_render_workspace(
    *,
    metadata: Metadata,
    main_package: Package,
    target: str,
    rpm_name: str,
    license_files: list[str],
    doc_files: list[str],
    distconf: Rust2RpmConf,
    feature_flags: FeatureFlags,
    rpmautospec: bool,
    auto_changelog_entry: bool,
    date: Optional[time.struct_time] = None,
    packager: Optional[str],
) -> str:
    template = spec_file_template("workspace.spec")

    # enforce consistent ordering of binaries
    binaries = [*metadata.get_binaries()]
    binaries.sort()

    is_cdylib = metadata.is_cdylib()

    rpm_version = Version.parse(main_package.version).to_rpm()
    rpm_description = main_package.get_description()
    rpm_summary = main_package.get_summary()

    generator_version = __version__.split(".")[0]

    buildrequires = rpm.workspace_buildrequires(metadata, feature_flags, False)
    test_requires = set.difference(
        rpm.workspace_buildrequires(metadata, feature_flags, True),
        rpm.workspace_buildrequires(metadata, feature_flags, False),
    )

    rpm_buildrequires = list(sorted(buildrequires))
    rpm_test_requires = list(sorted(test_requires))

    if any(package_uses_rust_1_60_feature_syntax(package.features) for package in metadata.packages):
        rust_packaging_dep = "cargo-rpm-macros >= 24"
    else:
        rust_packaging_dep = "rust-packaging >= 21"

    if feature_flags.all_features or distconf.all_features:
        cargo_args = " -a"
    elif distconf.enabled_features:
        # list of features is already sorted when parsing the configuration file
        cargo_args = f" -f {','.join(distconf.enabled_features)}"
    else:
        cargo_args = ""

    license_strs = {package.license for package in metadata.packages}
    if len(license_strs) == 1:
        rpm_license_tag, rpm_license_comments = translate_license(target, list(license_strs)[0])
    else:
        is_composite = (
            lambda x: (" " in x and " WITH " not in x and "(" not in x and ")" not in x) or "(" in x or ")" in x
        )
        license_strs = [
            license_str if not is_composite(license_str) else f"({license_str})"
            for license_str in license_strs
            if license_str is not None
        ]
        license_strs.sort()
        license_str = " AND ".join(license_strs)
        rpm_license_tag, rpm_license_comments = translate_license(target, license_str)

    template_args_common = {
        # Parameters specific to rust2rpm
        "rust2rpm_version": generator_version,
        "rust2rpm_target": target,
        "rust_packaging_dep": rust_packaging_dep,
        # Parameters for RPM package metadata
        "rpm_name": rpm_name,
        "rpm_version": rpm_version,
        "rpm_summary": distconf.summary or rpm_summary,
        "rpm_description": rpm_description,
        "rpm_license": rpm_license_tag,
        "rpm_license_comments": rpm_license_comments,
        "rpm_buildrequires": rpm_buildrequires,
        "rpm_test_requires": rpm_test_requires,
        "rpm_license_files": license_files,
        "rpm_doc_files": doc_files,
        "rpm_binary_names": binaries,
        "rpm_cdylib_package": is_cdylib,
        # Parameters derived from rust2rpm.conf
        "conf_buildrequires": distconf.buildrequires,
        "conf_test_requires": distconf.testrequires,
        "conf_bin_requires": distconf.bin_requires,
        # Parameters derived from command-line flags
        "cargo_args": cargo_args,
        "use_rpmautospec": rpmautospec,
        "make_changelog_entry": auto_changelog_entry,
    }

    match target:
        case "fedora":
            template_args_target = template_args_fedora(date, packager, rpmautospec)
        case "mageia":
            template_args_target = template_args_mageia(date, packager)
        case "opensuse":
            template_args_target = template_args_opensuse(date, packager)
        case "plain":
            template_args_target = template_args_plain(date, packager)
        case _:
            raise ValueError(f"Unknown target {target!r} (this should never happen)")

    spec_contents = template.render(
        **template_args_common,
        **template_args_target,
    )

    if not spec_contents.endswith("\n"):
        spec_contents += "\n"

    return spec_contents
